import shutil
import tempfile
from unittest import TestCase
import numpy as np
from types import SimpleNamespace
from astropy.io import fits as pyfits

from dkist_processing_pac import CUModel, TelescopeModel, FittingFramework, generic
from dkist_processing_pac.utils import gen_fake_data

RTOL=1e-6
ATOL=1e-6

class TestGenerateParameters(TestCase):

    def setUp(self):
        self.mode_opts = generic.init_fit_mode('baseline', 666, False)
        self.nummod = 10

    def get_number_in_parameter_range(self, parameter):

        num = (self.mode_opts[parameter]['max'] - self.mode_opts[parameter]['min']) * \
               np.random.random() + self.mode_opts[parameter]['min']

        if np.isnan(num):
            num = np.random.random()

        return num

    def test_correct(self):

        TM = TelescopeModel.TelescopeModel([90],[0],[45])
        TM.x34 = self.get_number_in_parameter_range('x34')
        TM.t34 = self.get_number_in_parameter_range('t34')
        TM.x56 = self.get_number_in_parameter_range('x56')
        TM.t56 = self.get_number_in_parameter_range('t56')
        CS = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                         np.array([False]),  np.array([57530]))

        params = FittingFramework.generate_parameters([100], TM, CS, self.mode_opts, self.nummod, fit_TM=False)
        d = params.valuesdict()
        self.assertTrue('x34' in d.keys())
        self.assertTrue('t34' in d.keys())
        self.assertTrue('x56' in d.keys())
        self.assertTrue('t56' in d.keys())
        self.assertTrue('ret0h_CS00' in d.keys())
        self.assertTrue('dreth_CS00' in d.keys())
        self.assertTrue('ret045_CS00' in d.keys())
        self.assertTrue('dret45_CS00' in d.keys())
        self.assertTrue('ret0r_CS00' in d.keys())
        self.assertTrue('dretr_CS00' in d.keys())
        self.assertTrue('t_pol_CS00' in d.keys())
        self.assertTrue('t_ret_CS00' in d.keys())
        self.assertTrue('I_sys_CS00' in d.keys())
        self.assertTrue('Q_in' in d.keys())
        self.assertTrue('U_in' in d.keys())
        self.assertTrue('V_in' in d.keys())
        self.assertEqual(d['x34'], TM.x34, msg='x34')
        self.assertEqual(d['t34'], TM.t34, msg='t34')
        self.assertEqual(d['x56'], TM.x56, msg='x56')
        self.assertEqual(d['t56'], TM.t56, msg='t56')
        self.assertEqual(d['ret0h_CS00'], self.mode_opts['ret0h']['value'])
        self.assertEqual(d['dreth_CS00'], self.mode_opts['dreth']['value'])
        self.assertEqual(d['ret045_CS00'], self.mode_opts['ret045']['value'])
        self.assertEqual(d['dret45_CS00'], self.mode_opts['dret45']['value'])
        self.assertEqual(d['ret0r_CS00'], self.mode_opts['ret0r']['value'])
        self.assertEqual(d['dretr_CS00'], self.mode_opts['dretr']['value'])
        self.assertEqual(d['t_pol_CS00'], self.mode_opts['t_pol']['value'])
        self.assertEqual(d['t_ret_CS00'], self.mode_opts['t_ret']['value'])
        self.assertEqual(d['I_sys_CS00'], 100)
        self.assertEqual(d['Q_in'], self.mode_opts['Q_in']['value'])
        self.assertEqual(d['U_in'], self.mode_opts['U_in']['value'])
        self.assertEqual(d['V_in'], self.mode_opts['V_in']['value'])

    def test_TM_true(self):

        TM = TelescopeModel.TelescopeModel([90],[0],[45])
        CS = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                         np.array([False]), np.array([57530]))

        params = FittingFramework.generate_parameters([100], TM, CS, self.mode_opts, self.nummod, fit_TM=True)
        self.assertTrue(params['x34'].vary)
        self.assertTrue(params['t34'].vary)
        self.assertTrue(params['x56'].vary)
        self.assertTrue(params['t56'].vary)

    def test_TM_false(self):

        TM = TelescopeModel.TelescopeModel([90],[0],[45])
        CS = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                         np.array([False]), np.array([57530]))

        params = FittingFramework.generate_parameters([100], TM, CS, self.mode_opts, self.nummod, fit_TM=False)
        self.assertFalse(params['x34'].vary)
        self.assertFalse(params['t34'].vary)
        self.assertFalse(params['x56'].vary)
        self.assertFalse(params['t56'].vary)

    def test_out_of_range_TM_pars(self):
        """We only need to test TM pars because the CM pars are set directly from the fit recipe"""

        TM = TelescopeModel.TelescopeModel([90],[0],[45])
        TM.x34 = self.mode_opts['x34']['max'] + 1
        TM.t34 = self.mode_opts['t34']['min'] - 1
        TM.x56 = self.mode_opts['x56']['min'] - 1
        TM.t56 = self.mode_opts['t56']['max'] + 1
        CS = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                         np.array([False]), np.array([57530]))

        params = FittingFramework.generate_parameters([100], TM, CS, self.mode_opts, self.nummod, fit_TM=False)
        d = params.valuesdict()
        self.assertEqual(d['x34'], self.mode_opts['x34']['max'])
        self.assertEqual(d['t34'], self.mode_opts['t34']['min'])
        self.assertEqual(d['x56'], self.mode_opts['x56']['min'])
        self.assertEqual(d['t56'], self.mode_opts['t34']['max'])

    def test_multi_CS(self):

        TM = TelescopeModel.TelescopeModel([90],[0],[45])
        CS1 = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                          np.array([False]), np.array([57530]), ret_0_r=0.98, dret_r=0.75)
        CS2 = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                          np.array([False]), np.array([57560]), ret_0_h=1., dret_h=0.8)
        CS3 = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                          np.array([False]), np.array([57590]), ret_0_h=2., dret_h=0.9)
        CS = CS1 + CS2 + CS3

        params = FittingFramework.generate_parameters([100, 200, 300], TM, CS, self.mode_opts, self.nummod)
        d = params.valuesdict()
        self.assertTrue('ret0h_CS00' in d.keys())
        self.assertTrue('dreth_CS00' in d.keys())
        self.assertTrue('ret045_CS00' in d.keys())
        self.assertTrue('dret45_CS00' in d.keys())
        self.assertTrue('ret0r_CS00' in d.keys())
        self.assertTrue('dretr_CS00' in d.keys())
        self.assertTrue('ret0h_CS01' in d.keys())
        self.assertTrue('dreth_CS01' in d.keys())
        self.assertTrue('ret045_CS01' in d.keys())
        self.assertTrue('dret45_CS01' in d.keys())
        self.assertTrue('ret0r_CS01' in d.keys())
        self.assertTrue('dretr_CS01' in d.keys())
        self.assertTrue('ret0h_CS02' in d.keys())
        self.assertTrue('dreth_CS02' in d.keys())
        self.assertTrue('ret045_CS02' in d.keys())
        self.assertTrue('dret45_CS02' in d.keys())
        self.assertTrue('ret0r_CS02' in d.keys())
        self.assertTrue('dretr_CS02' in d.keys())

    def test_relative_I(self):

        TM = TelescopeModel.TelescopeModel([90],[0],[45])
        TM.x34 = self.get_number_in_parameter_range('x34')
        TM.t34 = self.get_number_in_parameter_range('t34')
        TM.x56 = self.get_number_in_parameter_range('x56')
        TM.t56 = self.get_number_in_parameter_range('t56')
        CS = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                         np.array([False]),  np.array([57530]))

        params = FittingFramework.generate_parameters([100], TM, CS, self.mode_opts, self.nummod)
        self.assertEqual(params['I_sys_CS00'].min, 100 * self.mode_opts['I_sys']['min'])
        self.assertEqual(params['I_sys_CS00'].max, 100 * self.mode_opts['I_sys']['max'])

    def test_modmat_correct_num(self):

        TM = TelescopeModel.TelescopeModel([90],[0],[45])
        TM.x34 = self.get_number_in_parameter_range('x34')
        TM.t34 = self.get_number_in_parameter_range('t34')
        TM.x56 = self.get_number_in_parameter_range('x56')
        TM.t56 = self.get_number_in_parameter_range('t56')
        CS = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                        np.array([False]),  np.array([57530]))

        params = FittingFramework.generate_parameters([100], TM, CS, self.mode_opts, self.nummod)
        d = params.valuesdict()

        for s in ['I', 'Q', 'U', 'V']:
            for i in range(self.nummod):
                with self.subTest(stokes=s, m=i):
                    self.assertTrue('modmat_{}{}'.format(i,s) in d.keys())

    def test_modmat_norange(self):

        TM = TelescopeModel.TelescopeModel([90],[0],[45])
        TM.x34 = self.get_number_in_parameter_range('x34')
        TM.t34 = self.get_number_in_parameter_range('t34')
        TM.x56 = self.get_number_in_parameter_range('x56')
        TM.t56 = self.get_number_in_parameter_range('t56')
        CS = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                        np.array([False]),  np.array([57530]))

        params = FittingFramework.generate_parameters([100], TM, CS, self.mode_opts, self.nummod)

        for s in ['I', 'Q', 'U', 'V']:
            for i in range(self.nummod):
                with self.subTest(stokes=s, m=i):
                    self.assertEqual(params['modmat_{}{}'.format(i,s)].min, -np.inf)
                    self.assertEqual(params['modmat_{}{}'.format(i,s)].max, np.inf)

    def test_modmat_yesrange(self):

        TM = TelescopeModel.TelescopeModel([90],[0],[45])
        TM.x34 = self.get_number_in_parameter_range('x34')
        TM.t34 = self.get_number_in_parameter_range('t34')
        TM.x56 = self.get_number_in_parameter_range('x56')
        TM.t56 = self.get_number_in_parameter_range('t56')
        CS = CUModel.CalibrationSequence(np.array([0]), np.array([0]), np.array([True]), np.array([True]),
                                        np.array([False]),  np.array([57530]))

        self.mode_opts['switches']['fit_bounds'] = True

        params = FittingFramework.generate_parameters([100], TM, CS, self.mode_opts, self.nummod)

        for s in ['Q', 'U', 'V']:
            for i in range(self.nummod):
                with self.subTest(stokes=s, m=i):
                    self.assertEqual(params['modmat_{}{}'.format(i,s)].min, -1.5)
                    self.assertEqual(params['modmat_{}{}'.format(i,s)].max, 1.5)

        for i in range(self.nummod):
            with self.subTest(m=i):
                self.assertEqual(params['modmat_{}I'.format(i)].min, 0.5)
                self.assertEqual(params['modmat_{}I'.format(i)].max, 1.5)

class TestFitModulationMatrix(TestCase):

    def test_by_hand(self):
        TM = SimpleNamespace(M12=np.diag(np.ones(4)))
        TM.TM = np.dstack((np.diag(np.ones(4)) * 2, np.diag(np.ones(4)))).T
        CS = SimpleNamespace(s_in=np.array([1,0,0,0.]))
        CS.CM = np.dstack((np.ones((4,4)), np.diag(np.ones(4)) * 3)).T
        truth = np.array([[8, 0, 0, 0.],
                          [2, 2, 2, 2,],
                          [3, 0, 0, 0]])
        truth /= truth[0,0]
        I_cal = truth @ (TM.TM @ CS.CM @ TM.M12 @ CS.s_in).T
        S = (TM.TM @ CS.CM @ TM.M12 @ CS.s_in).T

        O = FittingFramework.fit_modulation_matrix(I_cal, S)
        np.testing.assert_allclose(truth, O, rtol=RTOL, atol=ATOL)

    def test_random(self):
        TM = SimpleNamespace(M12=np.diag(np.ones(4)), TM=np.random.random((45,4,4)))
        CS = SimpleNamespace(s_in=np.array([1,0,0,0.]), CM=np.random.random((45,4,4)))
        truth = np.random.random((10,4))
        truth /= truth[0,0]
        I_cal = truth @ (TM.TM @ CS.CM @ TM.M12 @ CS.s_in).T
        S = (TM.TM @ CS.CM @ TM.M12 @ CS.s_in).T

        O = FittingFramework.fit_modulation_matrix(I_cal, S)
        np.testing.assert_allclose(truth, O, rtol=RTOL, atol=ATOL)

class TestGenerateModelI(TestCase):

    def test_by_hand(self):
        TM = SimpleNamespace(M12=np.diag(np.ones(4)))
        TM.TM = np.dstack((np.diag(np.ones(4)) * 2, np.diag(np.ones(4)))).T
        CS = SimpleNamespace(s_in=np.array([1,0,0,0.]))
        CS.CM = np.dstack((np.ones((4,4)), np.diag(np.ones(4)) * 3)).T
        O = np.array([[1, 0, 0, 0.],
                      [2, 2, 2, 2,],
                      [3, 0, 0, 0]])
        truth = np.array([[2, 3],
                          [16,6],
                          [6, 9]])
        S = (TM.TM @ CS.CM @ TM.M12 @ CS.s_in).T

        I = FittingFramework.generate_model_I(O, S)
        np.testing.assert_allclose(truth, I, rtol=RTOL, atol=ATOL)

    def test_random(self):
        TM = SimpleNamespace(M12=np.diag(np.ones(4)), TM=np.random.random((45,4,4)))
        CS = SimpleNamespace(s_in=np.array([1,0,0,0.]), CM=np.random.random((45,4,4)))
        O = np.random.random((10,4))
        truth = O @ (TM.TM @ CS.CM @ TM.M12 @ CS.s_in).T
        S = (TM.TM @ CS.CM @ TM.M12 @ CS.s_in).T

        I = FittingFramework.generate_model_I(O, S)
        np.testing.assert_allclose(I, truth, rtol=RTOL, atol=ATOL)

class TestNaNValues(TestCase):

    def setUp(self) -> None:
        self.SoCC_dir = tempfile.mkdtemp()
        self.num_pos = 5
        gen_fake_data.SoCC_multi_day(self.SoCC_dir, numdays=3, shape=(1, 1, self.num_pos), DHS=False, CS_name='may_CS')

        self.CU_NaN_pos = np.random.choice(range(self.num_pos))
        self.TM_NaN_pos = np.random.choice(range(self.num_pos))
        for s, pos in zip([0, 2], [self.CU_NaN_pos, self.TM_NaN_pos]):
            step = np.random.choice(range(10)) + 2  # 2 comes from initial dark and clear
            mod = np.random.choice(range(10))
            hdl = pyfits.open('{:}/day{:}/VISP.C{:03n}.FITS'.format(self.SoCC_dir, s, step), mode='update')
            hdl[mod + 1].data[0, 0, pos] = np.nan
            hdl.flush()
            hdl.close()

        return

    def tearDown(self) -> None:
        self.CMP.cleanup()
        self.TMP.cleanup()
        shutil.rmtree(self.SoCC_dir)

    def test_CU_NaN(self):
        self.CMP, self.TMP, _ = FittingFramework.run_fits(['{}/day0'.format(self.SoCC_dir)], fit_TM=False,
                                                          telescope_db=generic.get_default_telescope_db())

        nan_list = list(self.CMP[0, 0, self.CU_NaN_pos].values())
        truth = np.zeros(len(nan_list))
        truth[-self.CMP.nummod * 4::4] = 1. / self.CMP.nummod
        np.testing.assert_equal(nan_list, truth)

    def test_CU_non_NaN(self):
        self.CMP, self.TMP, _ = FittingFramework.run_fits(['{}/day0'.format(self.SoCC_dir)], fit_TM=False,
                                                          telescope_db=generic.get_default_telescope_db())

        pos = (self.CU_NaN_pos + 1) % self.num_pos # guaranteed not to be the NaN position
        nan_list = list(self.CMP[0, 0, pos].values())
        truth = np.zeros(len(nan_list))
        truth[-self.CMP.nummod * 4::4] = 1. / self.CMP.nummod
        self.assertTrue(np.any(np.not_equal(nan_list, truth)))

    def test_TM_NaN(self):
        dirlist = ['{}/day{}'.format(self.SoCC_dir, i) for i in range(3)]
        self.CMP, self.TMP, _ = FittingFramework.run_fits(dirlist, fit_TM=True, threads=2,
                                                          telescope_db=generic.get_default_telescope_db())

        dummy_TM = np.array([0., 1., 0., 1., 0., 1.])
        np.testing.assert_equal(self.TMP[0, 0, self.CU_NaN_pos], dummy_TM)
        np.testing.assert_equal(self.TMP[0, 0, self.TM_NaN_pos], dummy_TM)

    def test_TM_non_NaN(self):
        dirlist = ['{}/day{}'.format(self.SoCC_dir, i) for i in range(3)]
        self.CMP, self.TMP, _ = FittingFramework.run_fits(dirlist, fit_TM=True, threads=2,
                                                          telescope_db=generic.get_default_telescope_db())

        pos1 = (self.CU_NaN_pos + 1) % self.num_pos
        if pos1 == self.TM_NaN_pos:
            pos1 = (pos1 + 1) % self.num_pos
        pos2 = (self.TM_NaN_pos + 1) % self.num_pos
        if pos2 == self.CU_NaN_pos:
            pos2 = (pos2 + 1) % self.num_pos

        dummy_TM = np.array([0., 1., 0., 1., 0., 1.])
        self.assertTrue(np.any(np.not_equal(self.TMP[0, 0, pos1], dummy_TM)))
        self.assertTrue(np.any(np.not_equal(self.TMP[0, 0, pos2], dummy_TM)))



"""
Use setup.cfg to configure packaging
"""
from setuptools import setup

setup(use_scm_version=True)

#
# #!/usr/bin/env python
# # Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# import glob
# import os
# import sys
#
# from setuptools import setup, find_packages
#
# # Get some values from the setup.cfg
# try:
#     from ConfigParser import ConfigParser
# except ImportError:
#     from configparser import ConfigParser
#
# conf = ConfigParser()
# conf.read(['setup.cfg'])
# metadata = dict(conf.items('metadata'))
#
# PACKAGENAME = metadata.get('package_name', 'dkist-processing-pac')
# DESCRIPTION = metadata.get('description', 'PA&C Pipeline')
# AUTHOR = metadata.get('author', 'Arthur Eigenbrot')
# AUTHOR_EMAIL = metadata.get('author_email', '')
# LICENSE = metadata.get('license', 'unknown')
# URL = metadata.get('url', 'http://dkist.nso.edu')
#
# # order of priority for long_description:
# #   (1) set in setup.cfg,
# #   (2) load LONG_DESCRIPTION.rst,
# #   (3) load README.rst,
# #   (4) package docstring
# readme_glob = 'README*'
# _cfg_long_description = metadata.get('long_description', '')
# if _cfg_long_description:
#     LONG_DESCRIPTION = _cfg_long_description
#
# elif os.path.exists('LONG_DESCRIPTION.rst'):
#     with open('LONG_DESCRIPTION.rst') as f:
#         LONG_DESCRIPTION = f.read()
#
# elif len(glob.glob(readme_glob)) > 0:
#     with open(glob.glob(readme_glob)[0]) as f:
#         LONG_DESCRIPTION = f.read()
#
# else:
#     # Get the long description from the package's docstring
#     __import__(PACKAGENAME)
#     package = sys.modules[PACKAGENAME]
#     LONG_DESCRIPTION = package.__doc__
#
# version_file = "dkist_processing_pac/_version.py"
# with open(version_file, 'r') as f:
#     lines = f.readlines()
#     VERSION = lines[0].split(' = ')[-1].replace('"', '')
#
# # Treat everything in scripts except README* as a script to be installed
# scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))
#            if not os.path.basename(fname).startswith('README')]
#
#
# # Get configuration information from all of the various subpackages.
# package_info = {'packages': find_packages(),
#                 'package_data': {}}
#
# # Add the project-global data
# package_info['package_data'].setdefault(PACKAGENAME, [])
# package_info['package_data'][PACKAGENAME].append('data/*')
#
# # Define entry points for command-line scripts
# entry_points = {'console_scripts': []}
#
# entry_point_list = conf.items('entry_points')
# for entry_point in entry_point_list:
#     entry_points['console_scripts'].append('{0} = {1}'.format(entry_point[0],
#                                                               entry_point[1]))
# setup(name=PACKAGENAME,
#       version=VERSION,
#       description=DESCRIPTION,
#       scripts=scripts,
#       install_requires=metadata.get('install_requires', 'astropy').strip().split(),
#       author=AUTHOR,
#       author_email=AUTHOR_EMAIL,
#       license=LICENSE,
#       url=URL,
#       long_description=LONG_DESCRIPTION,
#       zip_safe=False,
#       use_2to3=False,
#       entry_points=entry_points,
#       include_package_data=True,
#       **package_info
# )



import os
from collections import defaultdict
from glob import glob
from typing import Dict, Union
from typing import List
from typing import Tuple

import numpy as np
from astropy.io import fits as pyfits
import pytest
from dkist_data_simulator.dataset import key_function
from dkist_data_simulator.spec122 import Spec122Dataset
from dkist_header_validator import spec122_validator

from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess

from dkist_processing_pac import Data, FittingFramework, generic, GenerateDemodMatrices
from dkist_processing_pac.DKISTDC.data import DCDrawer
from dkist_processing_pac.utils import gen_fake_data


class CalibrationSequenceDataset(Spec122Dataset):
    def __init__(
        self,
        array_shape: Tuple[int, ...],
        time_delta: float,
        instrument="visp",
    ):
        self.num_mod = 3

        # Make up a Calibration sequence. Mostly random except for two clears and two darks at start and end, which
        # we want to test
        self.pol_status = [
            "clear",
            "clear",
            "Polarizer",
            "Polarizer",
            "Polarizer",
            "clear",
            "clear",
        ]
        self.pol_theta = [0.0, 0.0, 60.0, 60.0, 120.0, 0.0, 0.0]
        self.ret_status = ["clear", "clear", "clear", "SAR", "clear", "clear", "clear"]
        self.ret_theta = [0.0, 0.0, 0.0, 45.0, 0.0, 0.0, 0.0]
        self.dark_status = [
            "DarkShutter",
            "FieldStop (5arcmin)",
            "FieldStop (5arcmin)",
            "FieldStop (5arcmin)",
            "FieldStop (5arcmin)",
            "FieldStop (5arcmin)",
            "DarkShutter",
        ]

        self.num_steps = len(self.pol_theta)
        dataset_shape = (self.num_steps * self.num_mod,) + array_shape[1:]
        super().__init__(dataset_shape, array_shape, time_delta, instrument=instrument)
        self.add_constant_key("DKIST004", "polcal")
        self.add_constant_key("WAVELNTH", 666.)

    @property
    def cs_step(self) -> int:
        return self.index // self.num_mod

    @key_function("VISP_011")
    def modstate(self, key: str) -> int:
        return (self.index % self.num_mod) + 1

    @key_function("VISP_010")
    def nummod(self, key: str) -> int:
        return self.num_mod

    @key_function("PAC__004")
    def polarizer_status(self, key: str) -> str:
        return self.pol_status[self.cs_step]

    @key_function("PAC__005")
    def polarizer_angle(self, key: str) -> str:
        return str(self.pol_theta[self.cs_step])

    @key_function("PAC__006")
    def retarter_status(self, key: str) -> str:
        return self.ret_status[self.cs_step]

    @key_function("PAC__007")
    def retarder_angle(self, key: str) -> str:
        return str(self.ret_theta[self.cs_step])

    @key_function("PAC__008")
    def gos_level3_status(self, key: str) -> str:
        return self.dark_status[self.cs_step]

class InstAccess(L0FitsAccess):
    def __init__(self, hdu: Union[pyfits.ImageHDU, pyfits.PrimaryHDU, pyfits.CompImageHDU]):
        super().__init__(hdu, auto_squeeze=False)
        self.mod_state = self.header["VISP_011"]
        self.num_mod_states = self.header["VISP_010"]

@pytest.fixture(scope="session")
def cs_step_obj_dict() -> Dict[int, List[InstAccess]]:
    ds = CalibrationSequenceDataset(array_shape=(1, 2, 2), time_delta=2.0)
    header_list = [
        spec122_validator.validate_and_translate_to_214_l0(d.header(), return_type=pyfits.HDUList)[0].header
        for d in ds
    ]
    out_dict = dict()
    for n in range(7):
        hdu_list = []
        for m in range(3):
            hdu_list.append(pyfits.PrimaryHDU(data=np.ones((3, 4, 1)) * n + 100 * m, header=pyfits.Header(header_list.pop(0))))

        out_dict[n] = [InstAccess(h) for h in hdu_list]

    return out_dict

@pytest.fixture(scope="session")
def SoCC_dir(tmpdir_factory):
    SoCC_dir = tmpdir_factory.mktemp("SoCC")
    num_pos = 5
    gen_fake_data.SoCC_multi_day(SoCC_dir, numdays=1, shape=(1, 1, num_pos), DHS=False, CS_name='may_CS')

    return str(SoCC_dir / 'day0')

@pytest.fixture(scope="function")
def DC_SoCC(SoCC_dir):
    file_list = sorted(glob(os.path.join(SoCC_dir, '*.FITS')))
    obj_dict = defaultdict(list)
    for i, f in enumerate(file_list):
        hdl = pyfits.open(f)
        for h in hdl[1:]:
            t_head = spec122_validator.validate_and_translate_to_214_l0(h.header, return_type=pyfits.HDUList)[0].header
            obj_dict[i].append(InstAccess(pyfits.ImageHDU(data=h.data, header=t_head)))

    return obj_dict


def test_dkistdc_drawer(cs_step_obj_dict):

    D = DCDrawer(cs_step_obj_dict, remove_I_trend=False)
    assert D.nummod == 3
    assert D.numsteps == 7 - 2
    np.testing.assert_array_equal(D.pol_in, [False, True, True, True, False])
    np.testing.assert_array_equal(D.theta_pol_steps, [0.0, 60.0, 60.0, 120.0, 0.0])
    np.testing.assert_array_equal(D.ret_in, [False, False, True, False, False])
    np.testing.assert_array_equal(D.theta_ret_steps, [0.0, 0.0, 45.0, 0.0, 0.0])
    np.testing.assert_array_equal(D.dark_in, [False, False, False, False, False])
    assert D.shape == (3, 4, 1)
    cc = np.ones((3, 5)) * np.arange(1, 6)[None, :] + 100 * np.arange(3)[:, None]
    for i in range(np.prod(D.shape)):
        np.testing.assert_array_equal(D[np.unravel_index(i, D.shape)], cc)

def test_dkistdc_drawer_with_darks(cs_step_obj_dict):

    D = DCDrawer(cs_step_obj_dict, skip_darks=False, remove_I_trend=False)
    assert D.nummod == 3
    assert D.numsteps == 7
    np.testing.assert_array_equal(D.pol_in, [False, False, True, True, True, False, False])
    np.testing.assert_array_equal(D.theta_pol_steps, [0.0, 0.0, 60.0, 60.0, 120.0, 0.0, 0.0])
    np.testing.assert_array_equal(D.ret_in, [False, False, False, True, False, False, False])
    np.testing.assert_array_equal(D.theta_ret_steps, [0.0, 0.0, 0.0, 45.0, 0.0, 0.0, 0.0])
    np.testing.assert_array_equal(D.dark_in, [True, False, False, False, False, False, True])
    assert D.shape == (3, 4, 1)
    cc = np.ones((3, 7)) * np.arange(7)[None, :] + 100 * np.arange(3)[:, None]
    for i in range(np.prod(D.shape)):
        np.testing.assert_array_equal(D[np.unravel_index(i, D.shape)], cc)

def test_dkistdc_drawer_I_trend(cs_step_obj_dict):

    D = DCDrawer(cs_step_obj_dict, remove_I_trend=True)
    assert D.shape == (3, 4, 1)
    cc = np.ones((3, 5)) * np.arange(1, 6)[None, :] + 100 * np.arange(3)[:, None]
    cc /= np.arange(1, 6) * 0.00970874 + 0.97087379
    for i in range(np.prod(D.shape)):
        np.testing.assert_allclose(D[np.unravel_index(i, D.shape)], cc)

def test_dresser(cs_step_obj_dict):

    D1 = DCDrawer(cs_step_obj_dict, skip_darks=False, remove_I_trend=False)
    D2 = DCDrawer(cs_step_obj_dict, skip_darks=True, remove_I_trend=False)
    DRSR = Data.Dresser()
    DRSR.add_drawer(D1)
    DRSR.add_drawer(D2)

    assert DRSR.nummod == 3
    assert DRSR.numsteps == 7 + 7 - 2
    np.testing.assert_array_equal(DRSR.pol_in,
                                  [False, False, True, True, True, False, False] + [False, True, True, True, False])
    np.testing.assert_array_equal(DRSR.theta_pol_steps,
                                  [0.0, 0.0, 60.0, 60.0, 120.0, 0.0, 0.0] + [0.0, 60.0, 60.0, 120.0, 0.0])
    np.testing.assert_array_equal(DRSR.ret_in,
                                  [False, False, False, True, False, False, False] + [False, False, True, False, False])
    np.testing.assert_array_equal(DRSR.theta_ret_steps,
                                  [0.0, 0.0, 0.0, 45.0, 0.0, 0.0, 0.0] + [0.0, 0.0, 45.0, 0.0, 0.0])
    np.testing.assert_array_equal(DRSR.dark_in,
                                  [True, False, False, False, False, False, True] + [False, False, False, False, False])
    assert DRSR.shape == (3, 4, 1)
    cc1 = np.ones((3, 7)) * np.arange(7)[None, :] + 100 * np.arange(3)[:, None]
    cc2 = np.ones((3, 5)) * np.arange(1, 6)[None, :] + 100 * np.arange(3)[:, None]
    cc = np.hstack([cc1, cc2])
    for i in range(np.prod(DRSR.shape)):
        np.testing.assert_array_equal(DRSR[np.unravel_index(i, DRSR.shape)][0], cc)

def test_same_dresser_as_non_DC(SoCC_dir, DC_SoCC):
    DRSR = Data.Dresser()
    DRSR.add_drawer(Data.Drawer(SoCC_dir))

    dc_DRSR = Data.Dresser()
    dc_DRSR.add_drawer(DCDrawer(DC_SoCC))

    for attr in ['instrument', 'nummod', 'numsteps', 'shape', 'theta_pol_steps', 'theta_ret_steps', 'pol_in',
                 'ret_in', 'date_bgn', 'date_end', 'wavelength', 'azimuth', 'elevation', 'table_angle',
                 'I_clear']:
        print(attr)
        if type(getattr(DRSR, attr)) is np.ndarray:
            np.testing.assert_equal(getattr(DRSR, attr), getattr(dc_DRSR, attr))
        else:
            assert getattr(DRSR, attr) == getattr(dc_DRSR, attr)

    for i in range(np.prod(DRSR.shape)):
        idx = np.unravel_index(i, DRSR.shape)
        np.testing.assert_array_equal(DRSR[idx][0], dc_DRSR[idx][0])
        np.testing.assert_array_equal(DRSR[idx][1], dc_DRSR[idx][1])

@pytest.mark.slow
def test_same_fit_and_demod_as_non_DC(SoCC_dir, DC_SoCC, tmp_path):
    CMP, TMP, TM = FittingFramework.run_fits([SoCC_dir], fit_TM=False, threads=2,
                                             telescope_db=generic.get_default_telescope_db())
    cmp_file = tmp_path / 'cmp.fits'
    CMP.writeto(cmp_file)

    DRSR = Data.Dresser()
    DRSR.add_drawer(DCDrawer(DC_SoCC))
    dc_CMP, dc_TMP, dc_TM = FittingFramework.run_core(DRSR, fit_TM=False, threads=2,
                                             telescope_db=generic.get_default_telescope_db())

    np.testing.assert_equal(CMP.CU_params, dc_CMP.CU_params)
    np.testing.assert_equal(TMP.TM_params, dc_TMP.TM_params)

    dmod_file = tmp_path / 'dmod.fits'
    GenerateDemodMatrices.main(SoCC_dir, str(cmp_file), str(dmod_file),
                               telescope_db=generic.get_default_telescope_db())
    dhdl = pyfits.open(dmod_file)
    dmod = dhdl[1].data

    dc_dmod = GenerateDemodMatrices.DC_main(DRSR, dc_CMP, telescope_db=generic.get_default_telescope_db())
    np.testing.assert_equal(dmod, dc_dmod)



import os
from unittest import TestCase
import numpy as np
from dkist_processing_pac.TelescopeModel import TelescopeModel
from dkist_processing_pac import Data

class TestInit(TestCase):

    def test_array_input(self):

        az = el = tab = np.arange(13.)
        TM = TelescopeModel(az, el, tab)
        self.assertTrue(type(TM.azimuth) is np.ndarray)
        self.assertTrue(type(TM.elevation) is np.ndarray)
        self.assertTrue(type(TM.table_angle) is np.ndarray)
        self.assertTrue(TM.azimuth.shape == (13,))
        self.assertTrue(TM.elevation.shape == (13,))
        self.assertTrue(TM.table_angle.shape == (13,))

    def test_single_input(self):

        TM = TelescopeModel(1.2, 3.4, 5.6)
        self.assertTrue(type(TM.azimuth) is np.ndarray)
        self.assertTrue(type(TM.elevation) is np.ndarray)
        self.assertTrue(type(TM.table_angle) is np.ndarray)
        self.assertTrue(TM.azimuth.shape == (1,))
        self.assertTrue(TM.elevation.shape == (1,))
        self.assertTrue(TM.table_angle.shape == (1,))

    def test_mirror_params(self):

        TM = TelescopeModel([90],[0],[45])
        TM.x34 = 1.
        TM.t34 = 2.
        TM.x56 = 3.
        TM.t56 = 4.
        TM.x12 = 5.
        TM.t12 = 6.
        self.assertEqual(TM.x34, 1)
        self.assertEqual(TM.t34, 2)
        self.assertEqual(TM.x56, 3)
        self.assertEqual(TM.t56, 4)
        self.assertEqual(TM.x12, 5)
        self.assertEqual(TM.t12, 6)

    def test_numsteps(self):

        size = np.random.randint(1, 1000)
        az = el = tab = np.arange(size)
        TM = TelescopeModel(az, el, tab)
        self.assertEqual(TM.numsteps, size)

    def test_unequal_lengths(self):

        az = np.arange(4)
        el = np.arange(5)
        tab = np.arange(6)
        with self.assertRaises(ValueError):
            TM = TelescopeModel(az, el, tab)

    def test_angle_conversion(self):

        TM = TelescopeModel(90, 60, 45)
        np.testing.assert_array_equal(TM.azimuth, np.array([np.pi/2.]))
        np.testing.assert_array_equal(TM.elevation, np.array([np.pi/3.]))
        np.testing.assert_array_equal(TM.table_angle, np.array([np.pi/4.]))

class TestAddition(TestCase):

    def test_concatenation(self):

        TM1 = TelescopeModel([1,2],[3,4],[5,6])
        TM2 = TelescopeModel([3,4],[5,6],[7,8])
        TM = TM1 + TM2
        np.testing.assert_array_equal(TM.azimuth, np.array([1,2,3,4]) * np.pi / 180)
        np.testing.assert_array_equal(TM.elevation, np.array([3,4,5,6]) * np.pi / 180)
        np.testing.assert_array_equal(TM.table_angle, np.array([5,6,7,8]) * np.pi / 180)

    def test_numsteps(self):

        TM1 = TelescopeModel([1,2],[3,4],[5,6])
        TM2 = TelescopeModel([3,4],[5,6],[7,8])
        TM = TM1 + TM2
        self.assertEqual(TM.numsteps, 4)

    def test_new_object(self):

        TM1 = TelescopeModel([1,2],[3,4],[5,6])
        TM2 = TelescopeModel([3,4],[5,6],[7,8])
        TM = TM1 + TM2
        self.assertIsNot(TM1, TM)
        self.assertIsNot(TM2, TM)

class TestComponents(TestCase):

    def setUp(self):

        self.TM = TelescopeModel([0,90.], [60, 75.], [90, 0.])
        self.TM.x12 = self.x12 = np.random.random() * (1.25 - 0.95) + 0.95
        self.TM.x34 = self.x34 = np.random.random() * (1.25 - 0.95) + 0.95
        self.TM.x56 = self.x56 = np.random.random() * (1.25 - 0.95) + 0.95
        self.TM.t12 = self.t12 = np.random.random() * (3.75 - 2.53) + 2.53
        self.TM.t34 = self.t34 = np.random.random() * (3.75 - 2.53) + 2.53
        self.TM.t56 = self.t56 = np.random.random() * (3.75 - 2.53) + 2.53

    def test_M12(self):
        np.testing.assert_array_equal(self.TM.M12, Data.mirror(self.x12, self.t12))
        self.assertEqual(self.TM.M12.shape, (4,4))

    def test_M34(self):
        np.testing.assert_array_equal(self.TM.M34, Data.mirror(self.x34, self.t34))
        self.assertEqual(self.TM.M34.shape, (4,4))

    def test_M56(self):
        np.testing.assert_array_equal(self.TM.M56, Data.mirror(self.x56, self.t56))
        self.assertEqual(self.TM.M56.shape, (4,4))

    def test_R23(self):
        np.testing.assert_array_equal(self.TM.R23, Data.rotation(-np.pi/2.))
        self.assertEqual(self.TM.R23.shape, (4,4))

    def test_R45(self):
        np.testing.assert_array_equal(self.TM.R45, np.transpose(np.dstack((Data.rotation(-np.pi/3.),
                                                                           Data.rotation(-75 * np.pi / 180))), (2,0,1)))
        self.assertEqual(self.TM.R45.shape, (2,4,4))

    def test_R67(self):
        np.testing.assert_array_equal(self.TM.R67, np.transpose(np.dstack((Data.rotation(0.),
                                                                           Data.rotation(np.pi))), (2,0,1)))
        self.assertEqual(self.TM.R67.shape, (2,4,4))

    def test_TM(self):
        np.testing.assert_array_equal(self.TM.TM, self.TM.R67 @ self.TM.M56 @ self.TM.R45 @ self.TM.M34 @ self.TM.R23)
        self.assertEqual(self.TM.TM.shape, (2,4,4))

class TestLoadDatabase(TestCase):

    def setUp(self):
        self.dbfile = 'tmpdb.txt'
        with open(self.dbfile, 'a') as f:
            f.write(str('{:15.6f}' + '{:8.2f}' + '{:7.3f}'*6 + '\n').format(55000,
                                                                            500,
                                                                            1, 10,
                                                                            3, 30,
                                                                            5, 50))
            f.write(str('{:15.6f}' + '{:8.2f}' + '{:7.3f}'*6 + '\n').format(56000,
                                                                            500,
                                                                            2, 20,
                                                                            4, 40,
                                                                            6, 60))
            f.write(str('{:15.6f}' + '{:8.2f}' + '{:7.3f}'*6 + '\n').format(55000,
                                                                            600,
                                                                            1.5, 15,
                                                                            3.5, 35,
                                                                            5.5, 55))
            f.write(str('{:15.6f}' + '{:8.2f}' + '{:7.3f}'*6 + '\n').format(56000,
                                                                            600,
                                                                            2.5, 25,
                                                                            4.5, 45,
                                                                            6.5, 65))

    def tearDown(self):
        os.remove(self.dbfile)

    def test_exact_match(self):

        TM = TelescopeModel(1,2,3)
        TM.load_from_database(self.dbfile, 55000, 500.)
        self.assertEqual(TM.x12, 1)
        self.assertEqual(TM.t12, 10)
        self.assertEqual(TM.x34, 3)
        self.assertEqual(TM.t34, 30)
        self.assertEqual(TM.x56, 5)
        self.assertEqual(TM.t56, 50)
        self.assertEqual(TM.wavelength, 500)
        self.assertEqual(TM.obstime, 55000)

    def test_wave_linear_interp(self):

        TM = TelescopeModel(1,2,3)
        TM.load_from_database(self.dbfile, 55000, 550, method='linear')
        self.assertEqual(TM.x12, 1.25)
        self.assertEqual(TM.t12, 12.5)
        self.assertEqual(TM.x34, 3.25)
        self.assertEqual(TM.t34, 32.5)
        self.assertEqual(TM.x56, 5.25)
        self.assertEqual(TM.t56, 52.5)
        self.assertEqual(TM.wavelength, 550)
        self.assertEqual(TM.obstime, 55000)

    def test_time_linear_interp(self):

        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 55500, 500, method='linear')
        self.assertEqual(TM.x12, 1.5)
        self.assertEqual(TM.t12, 15)
        self.assertEqual(TM.x34, 3.5)
        self.assertEqual(TM.t34, 35)
        self.assertEqual(TM.x56, 5.5)
        self.assertEqual(TM.t56, 55)
        self.assertEqual(TM.wavelength, 500)
        self.assertEqual(TM.obstime, 55500)

    def test_both_linear_interp(self):

        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 55500, 550, method='linear')
        self.assertEqual(TM.x12, 1.75)
        self.assertEqual(TM.t12, 17.5)
        self.assertEqual(TM.x34, 3.75)
        self.assertEqual(TM.t34, 37.5)
        self.assertEqual(TM.x56, 5.75)
        self.assertEqual(TM.t56, 57.5)
        self.assertEqual(TM.wavelength, 550)
        self.assertEqual(TM.obstime, 55500)

    def test_wave_nearest_interp(self):

        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 55000, 550, method='nearest')
        self.assertEqual(TM.x12, 1)
        self.assertEqual(TM.t12, 10)
        self.assertEqual(TM.x34, 3)
        self.assertEqual(TM.t34, 30)
        self.assertEqual(TM.x56, 5)
        self.assertEqual(TM.t56, 50)
        self.assertEqual(TM.wavelength, 550)
        self.assertEqual(TM.obstime, 55000)

    def test_time_nearest_interp(self):

        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 55501, 500, method='nearest')
        self.assertEqual(TM.x12, 2)
        self.assertEqual(TM.t12, 20)
        self.assertEqual(TM.x34, 4)
        self.assertEqual(TM.t34, 40)
        self.assertEqual(TM.x56, 6)
        self.assertEqual(TM.t56, 60)
        self.assertEqual(TM.wavelength, 500)
        self.assertEqual(TM.obstime, 55501)

    def test_both_nearest_interp(self):

        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 55500, 551, method='nearest')
        self.assertEqual(TM.x12, 1.5)
        self.assertEqual(TM.t12, 15)
        self.assertEqual(TM.x34, 3.5)
        self.assertEqual(TM.t34, 35)
        self.assertEqual(TM.x56, 5.5)
        self.assertEqual(TM.t56, 55)
        self.assertEqual(TM.wavelength, 551)
        self.assertEqual(TM.obstime, 55500)

    def test_wave_out_of_bounds_linear(self):
        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 55000, 700, method='linear')
        self.assertEqual(TM.x12, 1.5)
        self.assertEqual(TM.t12, 15)
        self.assertEqual(TM.x34, 3.5)
        self.assertEqual(TM.t34, 35)
        self.assertEqual(TM.x56, 5.5)
        self.assertEqual(TM.t56, 55)
        self.assertEqual(TM.wavelength, 700)
        self.assertEqual(TM.obstime, 55000)

    def test_time_out_of_bounds_linear(self):
        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 4, 500, method='linear')
        self.assertEqual(TM.x12, 1)
        self.assertEqual(TM.t12, 10)
        self.assertEqual(TM.x34, 3)
        self.assertEqual(TM.t34, 30)
        self.assertEqual(TM.x56, 5)
        self.assertEqual(TM.t56, 50)
        self.assertEqual(TM.wavelength, 500)
        self.assertEqual(TM.obstime, 4)

    def test_both_out_of_bounds_linear(self):
        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 1e7, 2, method='linear')
        self.assertEqual(TM.x12, 2)
        self.assertEqual(TM.t12, 20)
        self.assertEqual(TM.x34, 4)
        self.assertEqual(TM.t34, 40)
        self.assertEqual(TM.x56, 6)
        self.assertEqual(TM.t56, 60)
        self.assertEqual(TM.wavelength, 2)
        self.assertEqual(TM.obstime, 1e7)

    def test_wave_out_of_bounds_nearest(self):
        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 55000, 700, method='nearest')
        self.assertEqual(TM.x12, 1.5)
        self.assertEqual(TM.t12, 15)
        self.assertEqual(TM.x34, 3.5)
        self.assertEqual(TM.t34, 35)
        self.assertEqual(TM.x56, 5.5)
        self.assertEqual(TM.t56, 55)
        self.assertEqual(TM.wavelength, 700)
        self.assertEqual(TM.obstime, 55000)

    def test_time_out_of_bounds_nearest(self):
        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 4, 500, method='nearest')
        self.assertEqual(TM.x12, 1)
        self.assertEqual(TM.t12, 10)
        self.assertEqual(TM.x34, 3)
        self.assertEqual(TM.t34, 30)
        self.assertEqual(TM.x56, 5)
        self.assertEqual(TM.t56, 50)
        self.assertEqual(TM.wavelength, 500)
        self.assertEqual(TM.obstime, 4)

    def test_both_out_of_bounds_nearest(self):
        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 1e7, 2, method='nearest')
        self.assertEqual(TM.x12, 2)
        self.assertEqual(TM.t12, 20)
        self.assertEqual(TM.x34, 4)
        self.assertEqual(TM.t34, 40)
        self.assertEqual(TM.x56, 6)
        self.assertEqual(TM.t56, 60)
        self.assertEqual(TM.wavelength, 2)
        self.assertEqual(TM.obstime, 1e7)

    def test_wave_out_of_bounds_cubic(self):
        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 55000, 700, method='cubic')
        self.assertEqual(TM.x12, 1.5)
        self.assertEqual(TM.t12, 15)
        self.assertEqual(TM.x34, 3.5)
        self.assertEqual(TM.t34, 35)
        self.assertEqual(TM.x56, 5.5)
        self.assertEqual(TM.t56, 55)
        self.assertEqual(TM.wavelength, 700)
        self.assertEqual(TM.obstime, 55000)

    def test_time_out_of_bounds_cubic(self):
        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 4, 500, method='cubic')
        self.assertEqual(TM.x12, 1)
        self.assertEqual(TM.t12, 10)
        self.assertEqual(TM.x34, 3)
        self.assertEqual(TM.t34, 30)
        self.assertEqual(TM.x56, 5)
        self.assertEqual(TM.t56, 50)
        self.assertEqual(TM.wavelength, 500)
        self.assertEqual(TM.obstime, 4)

    def test_both_out_of_bounds_cubic(self):
        TM = TelescopeModel(1, 2, 3)
        TM.load_from_database(self.dbfile, 1e7, 2, method='cubic')
        self.assertEqual(TM.x12, 2)
        self.assertEqual(TM.t12, 20)
        self.assertEqual(TM.x34, 4)
        self.assertEqual(TM.t34, 40)
        self.assertEqual(TM.x56, 6)
        self.assertEqual(TM.t56, 60)
        self.assertEqual(TM.wavelength, 2)
        self.assertEqual(TM.obstime, 1e7)

class TestSaveDatabase(TestCase):

    def setUp(self):
        self.dbfile = 'tmpdb.txt'
        self.x12 = 1.2
        self.t12 = 12.3
        self.x34 = 3.4
        self.t34 = 34.5
        self.x56 = 5.6
        self.t56 = 56.7
        self.TM = TelescopeModel(1,2,3)
        self.TM.x12 = self.x12
        self.TM.t12 = self.t12
        self.TM.x34 = self.x34
        self.TM.t34 = self.t34
        self.TM.x56 = self.x56
        self.TM.t56 = self.t56
        with open(self.dbfile, 'a') as f:
            f.write(str('{:15.6f}' + '{:8.2f}' + '{:7.3f}'*6 + '\n').format(55000,
                                                                            500,
                                                                            1, 10,
                                                                            3, 30,
                                                                            5, 50))

    def tearDown(self):
        os.remove(self.dbfile)

    def test_save(self):

        truth = str('{:15.6f}' + '{:8.2f}' + '{:7.3f}'*6 + '\n').format(100,
                                                                        200,
                                                                        self.x12, self.t12,
                                                                        self.x34, self.t34,
                                                                        self.x56, self.t56)
        self.TM.save_to_database(self.dbfile, 100, 200)
        with open(self.dbfile, 'r') as f:
            lines = f.readlines()
        self.assertEqual(lines[-1], truth)

    def test_save_class_wave(self):

        truth = str('{:15.6f}' + '{:8.2f}' + '{:7.3f}'*6 + '\n').format(100,
                                                                        222,
                                                                        self.x12, self.t12,
                                                                        self.x34, self.t34,
                                                                        self.x56, self.t56)
        self.TM.wavelength = 222.
        self.TM.save_to_database(self.dbfile, obstime=100)
        with open(self.dbfile, 'r') as f:
            lines = f.readlines()
        self.assertEqual(lines[-1], truth)

    def test_save_class_time(self):

        truth = str('{:15.6f}' + '{:8.2f}' + '{:7.3f}'*6 + '\n').format(111,
                                                                        200,
                                                                        self.x12, self.t12,
                                                                        self.x34, self.t34,
                                                                        self.x56, self.t56)
        self.TM.obstime = 111
        self.TM.save_to_database(self.dbfile, obswave=200)
        with open(self.dbfile, 'r') as f:
            lines = f.readlines()
        self.assertEqual(lines[-1], truth)

    def test_save_class_both(self):

        truth = str('{:15.6f}' + '{:8.2f}' + '{:7.3f}'*6 + '\n').format(111,
                                                                        222,
                                                                        self.x12, self.t12,
                                                                        self.x34, self.t34,
                                                                        self.x56, self.t56)
        self.TM.obstime = 111
        self.TM.wavelength = 222
        self.TM.save_to_database(self.dbfile)
        with open(self.dbfile, 'r') as f:
            lines = f.readlines()
        self.assertEqual(lines[-1], truth)

    def test_save_no_wave(self):

        self.TM.obstime = 111
        with self.assertRaises(ValueError):
            self.TM.save_to_database(self.dbfile)

    def test_save_no_time(self):

        with self.assertRaises(ValueError):
            self.TM.save_to_database(self.dbfile, obswave=200)

